/*
 * Copyright (C) 1999 Yasuhiro Ohara
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.  
 */

#ifndef OSPF6_LSA_H
#define OSPF6_LSA_H

#define THIS_IS_REFRESH  ((struct thread *)0xffffffff)

/* LSA definition */

/* Type */
#define LST_V2ROUTER_LSA               1
#define LST_V2NETWORK_LSA              2
#define LST_V2TYPE3_SUMMARY_LSA        3 /* routes to network */
#define LST_V2TYPE4_SUMMARY_LSA        4 /* routes to ASBR */
#define LST_V2AS_EXTERNAL_LSA          5

#define LST_ROUTER_LSA              0x2001
#define LST_NETWORK_LSA             0x2002
#define LST_INTER_AREA_PREFIX_LSA   0x2003
#define LST_INTER_AREA_ROUTER_LSA   0x2004
#define LST_AS_EXTERNAL_LSA         0x4005
#define LST_GROUP_MEMBERSHIP_LSA    0x2006
#define LST_TYPE_7_LSA              0x2007
#define LST_LINK_LSA                0x0008
#define LST_INTRA_AREA_PREFIX_LSA   0x2009

/* lsa scope */
#define SCOPE_MASK       0x6000
#define SCOPE_LINKLOCAL  0x0000
#define SCOPE_AREA       0x2000
#define SCOPE_AS         0x4000
#define SCOPE_RESERVED   0x6000
#define GET_LSASCOPE(x) ((ntohs(x)) & SCOPE_MASK)

/* NOTE that all lsa is left NETWORK BYTE ORDER */

struct router_lsa
{
  unsigned char rlsa_bits;
  unsigned char rlsa_options[3];
  /* followed by router_lsd(s) */
};
#define ROUTER_LSA_BIT_B     (1 << 0)
#define ROUTER_LSA_BIT_E     (1 << 1)
#define ROUTER_LSA_BIT_V     (1 << 2)
#define ROUTER_LSA_BIT_W     (1 << 3)

#define ROUTER_LSA_SET(x,y)    ((x)->rlsa_bits |=  (y))
#define ROUTER_LSA_ISSET(x,y)  ((x)->rlsa_bits &   (y))
#define ROUTER_LSA_CLEAR(x,y)  ((x)->rlsa_bits &= ~(y))

struct router_lsd
{
  unsigned char  rlsd_type;
  unsigned char  rlsd_reserved;
  unsigned short rlsd_metric;                /* output cost */
  unsigned long  rlsd_interface_id;
  unsigned long  rlsd_neighbor_interface_id;
  unsigned long  rlsd_neighbor_router_id;
};

#define LSDT_POINTTOPOINT       1
#define LSDT_TRANSIT_NETWORK    2
#define LSDT_STUB_NETWORK       3
#define LSDT_VIRTUAL_LINK       4

struct network_lsa
{
  unsigned char nlsa_reserved;
  unsigned char nlsa_options[3];
  /* followed by router_id(s) */
};

struct link_lsa
{
  unsigned char   llsa_rtr_pri;
  unsigned char   llsa_options[3];
  struct in6_addr llsa_linklocal;
  unsigned long   llsa_prefix_num;
  /* followed by prefix(es) */
};

struct intra_area_prefix_lsa
{
  unsigned short intra_prefix_num;
  unsigned short intra_prefix_refer_lstype;
  unsigned long  intra_prefix_refer_lsid;
  unsigned long  intra_prefix_refer_advrtr;
};

/* new */
struct ospf6_lsa_hdr
{
  unsigned short lsh_age;      /* LS age */
  unsigned short lsh_type;     /* LS type */
  unsigned long  lsh_id;       /* Link State ID */
  unsigned long  lsh_advrtr;   /* Advertising Router */
  unsigned long  lsh_seqnum;   /* LS sequence number */
  unsigned short lsh_cksum;    /* LS checksum */
  unsigned short lsh_len;      /* length */
};

#define LSH_NEXT(x) ((x) + 1)
#define LSA_NEXT(x) ((struct ospf6_lsa_hdr *) \
                       ((char *)(x) + ntohs ((x)->lsh_len)))
#define lsa_issame(x,y) ((x)->lsh_type == (y)->lsh_type && \
                         (x)->lsh_id == (y)->lsh_id && \
                         (x)->lsh_advrtr == (y)->lsh_advrtr)

struct ospf6_lsa
{
  unsigned long          lock;      /* reference counter */
  struct ospf6_lsa_hdr  *lsa_hdr;
  void                  *scope;     /* pointer of scoped data structure */
  unsigned char          flags;     /* use this to decide ack type */
  unsigned long          birth;     /* tv_sec when LS age 0 */
  unsigned long          installed; /* tv_sec when installed */
  struct thread         *expire;
  struct thread         *refresh;   /* For self-originated LSA */
  struct neighbor       *from;      /* from which neighbor */
  list                  summary_nbr;
  list                  request_nbr;
  list                  retrans_nbr;
  list                  delayed_ack_if;
};
#define OSPF6_LSA_FLOODBACK   (1 << 0)
#define OSPF6_LSA_DUPLICATE   (1 << 1)
#define OSPF6_LSA_IMPLIEDACK  (1 << 2)

/* Function Prototypes */
int past_min_ls_interval (struct ospf6_lsa *);
int which_is_more_recent (struct ospf6_lsa *, struct ospf6_lsa *);
void originating_lsa (struct ospf6_lsa *);
void construct_router_lsa (struct area *);
void construct_network_lsa (struct ospf6_if *);
void construct_link_lsa (struct ospf6_if *);
void construct_intra_prefix_lsa (struct ospf6_if *);
int show_router_lsa (struct vty *, void *);
int show_network_lsa (struct vty *, void *);
int show_link_lsa (struct vty *, void *);
int show_intra_prefix_lsa (struct vty *, void *);
int vty_lsa (struct vty *, struct ospf6_lsa *);

struct router_lsd *
get_router_lsd (rtr_id_t, struct ospf6_lsa *);
unsigned long get_ifindex_to_router (rtr_id_t, struct ospf6_lsa *);
void get_referencing_lsa (list, struct ospf6_lsa *);
int is_self_originated (struct ospf6_lsa *);
void update_ls_seqnum (struct ospf6_lsa *);
void reconstruct_lsa (struct ospf6_lsa *);

void ospf6_lsa_lock (struct ospf6_lsa *);
void ospf6_lsa_unlock (struct ospf6_lsa *);
int ospf6_lsa_expire (struct thread *);
int ospf6_lsa_refresh (struct thread *);
unsigned short ospf6_age_current (struct ospf6_lsa *);
void ospf6_age_update_to_send (struct ospf6_lsa *, struct ospf6_if *);
struct ospf6_lsa_hdr *make_ospf6_lsa_data (struct ospf6_lsa_hdr *, int);
struct ospf6_lsa *make_ospf6_lsa (struct ospf6_lsa_hdr *);
unsigned short ospf6_lsa_get_type (struct ospf6_lsa *);
unsigned short ospf6_lsa_get_scope_type (unsigned short);
void ospf6_lsa_clear_flag (struct ospf6_lsa *);
void ospf6_lsa_set_flag (struct ospf6_lsa *, unsigned char);
int ospf6_lsa_test_flag (struct ospf6_lsa *, unsigned char);
int ospf6_lsa_issame (struct ospf6_lsa_hdr *, struct ospf6_lsa_hdr *);

#endif /* OSPF6_LSA_H */

